/*
 * Name: Solar Tracker EspCode.ino (test version)
 * Author: Sorin Liviu Jurj
 * 
 * Description:
 * This software will implement different testing testing techniques to test 
 * the software functionality and features. We will tackle each part of the 
 * code individually and attempt to show all the possible break points and 
 * try to detect possible fault errors using white box testing techniques.
 *    This testing is based on the AUnit arduino library which is a port of
 * ArduinoUnit and Google Test testing platforms.
 */
 
#include <Stepper.h> 
#include <math.h> 
#include <SoftwareSerial.h>

// Library for testing 
#include <AUnit.h>
using aunit::TestRunner;      // name space for aunit to prevent name clashing 

/*
 * Please uncomment FIELD MODE and comment TEST MODE in order to receive analog readings
 */
#define TEST_MODE 1
// #define FIELD_MODE 0

uint8_t motorStephor = 200, motorStepver = 200;
uint8_t motor1hor = 4, motor2hor = 5, motor3hor = 6, motor4hor = 7;
uint8_t motor1ver = 8, motor2ver = 9, motor3ver = 10, motor4ver = 11;

int average, h=20, v=5; 
int ltsensor, rtsensor, rdsensor, ldsensor; 
int sen=50; 

int dil, dit, dir, did, diff, diff2; 
int pup = 12, pdown = 13; //lower switch

// definitions of stepper motors 
Stepper horStep (motorStephor, motor1hor, motor2hor, motor3hor, motor4hor);
Stepper verStep (motorStepver, motor1ver, motor2ver, motor3ver, motor4ver);

// serial communication with ESP
SoftwareSerial mySerial(2, 3); // RX, TX


//---------------------------------------------
//---------------------------------------------
//                TESTS
//---------------------------------------------
//--------------------------------------------- 
/*
 * We will run a continious test on the LTSensor,
 * LDSensor, RTSensor, and RDSensor data values to 
 * make that they are within the specified range.
 * This can be a source of error if the values are 
 * outside the specified range, as this can lead to 
 * errora and unpredictable movement of the stepper
 * motors.
 */
testing(sensorData)
{
  assertLessOrEqual(ltsensor, 1024);
  assertLessOrEqual(rtsensor, 1024);
  assertLessOrEqual(rdsensor, 1024);
  assertLessOrEqual(ldsensor, 1024);
}

 
/*
 * Main Program Setup
 */
void setup ()
{
  /*
   * The values for the intial speed of the stepper motors are hard coded, so there is little room for error.
   * In a setup where they are selected dynamically, we will have to run a test to make sure that they are 
   * within a specified range.
   */
  horStep.setSpeed (30); 
  verStep.setSpeed (10); 
  
  Serial.begin(9600);
  mySerial.begin(9600);
  
  /*
   * This while statement make sure that the serial connection is completely setup and running before, we can 
   * continue. In some cases, the serial communication msy not start fast enough. This will lead to data 
   * corruption and error if we are trying to receive or send data when the serial connection is not yet completly 
   * ready.
   */
  while (!Serial && !mySerial);
  
  

  pinMode(pup, INPUT);
  pinMode(pdown, INPUT);
}


// this is a long variable that holds the last time we set the sent analog data to the server. We use it to 
// control the frequency of data transmission to the server. Sending that continously can lead to over tasking
// the server script that processed the received data, and can also cause framing errors in serial communication.
long lastSendTime = 0;
/*
 * Main Program Loop
 */
void loop ()
{
  // send data to server every 3 seconds
  if(millis()-lastSendTime > 3000)
  {
    Data_Send();
    lastSendTime = millis();
  }
  
  do
  {
    /*
     * if the device is set in field mode we read control data form the ananlog sensors, But if it is
     * operating in test mode (receiving control instructions from our mqtt server), we do not need to
     * read data from analog sensors. This selection process is necessary to ensure that the data being 
     * used to control the motors is solely from one source.
     */
#ifdef FIELD_MODE
    ltsensor = analogRead(1)*1.022; 
    rtsensor = analogRead(2)*1.007;
    ldsensor = analogRead(3);
    rdsensor = analogRead(4)*1.013;
#endif
  
#ifdef TEST_MODE
     Receive_Data();
#endif
    average= (ltsensor + ldsensor + rtsensor + rdsensor)/4; 

    dit = (ltsensor + rtsensor)/2; 
    did = (ldsensor + rdsensor)/2; 
    diff =(dit - did); 
    delay (50); 
         
   if ((pup==HIGH)&&(average<=8)|| (pdown==HIGH)&&(average<=8)) //If the value of the average of the sensors is equal or less than 8 and the switches have the range
  
   mov(); //mov function
  
  }
  
  while ((pup==HIGH)&&(average<=8)||(pdown==HIGH)&&(average<=8));
  
  if (-1*sen > diff || diff > sen) //If the measured difference between the set of sensors is greater or less than the sensitivity value
  {
    if(dit < did) //If the mean value of the above sensors is smaller than the bottom sensors
    {
      if (pdown==HIGH)
      {
        verStep.step (0); //Stop vertical motor
        delay (10);
      }
      else if (pdown==LOW)
      {
        verStep.step (v); //Turn motor up
        delay (50);
      }
    }
    
    else if(dit > did) //If the average value of bottom sensors is smaller than the above sensors
    {
      if (pup==HIGH)
      {
        verStep.step (0); //Stop vertical motor
        delay (10);
      }
    
      else if (pup==LOW)
      {
        verStep.step (-v); //Turn motor down
        delay (50);
      }
    }
    
    else //any other case
    {
      verStep.step (0); //Stop vertical motor
      delay (10);
    }
  
  }
  
  delay (10);
  
#ifdef FIELD_MODE
    ltsensor = analogRead(1)*1.022; 
    rtsensor = analogRead(2)*1.007;
    ldsensor = analogRead(3);
    rdsensor = analogRead(4)*1.013;
#endif
  
#ifdef TEST_MODE
    Receive_Data();
#endif
  
  dil = (ltsensor + ldsensor)/2; 
  dir = (rtsensor + rdsensor)/2; 
  diff2 = (dil - dir); 
  delay (50);
  
 if (-1*sen > diff2 || diff2 > sen) //If the measured difference between the set of sensors is greater or less than the sensitivity value
  {
    if(dil < dir) //If the average of the left sensor is smaller than the right sensor
    {
      horStep.step (h); //Turn motor right
      delay (10);
    }
    else if(dil > dir) //If the average of the left sensor is larger than the right sensor
    {
      horStep.step(-h); //Turn motor left
      delay (10);
    }
    else //any other case
    {
      horStep.step(0); //Stop horizontal motor
      delay (10);
    }
  }
  delay(10);
}



/*
 * Move Function 
 */
void mov ()
{

  if (pup==HIGH)
  {
    verStep.step (72); //Turn 72 steps up (are the steps to change position once hide the sun)
    delay (50);
  }
  
  else if (pdown==HIGH)
  {
    verStep.step (-72); //Turn 72 steps down
    delay (50); 
  }
  
  delay (10);
}


void Data_Send()
{
  // get data from sensors
  int d1 = analogRead(1)*1.022; 
  int d2 = analogRead(2)*1.007;
  int d3 = analogRead(3);
  int d4 = analogRead(4)*1.013;
    
  Serial.println("Sending to ESP");
  mySerial.print(d1); 
  mySerial.print('.');
  mySerial.print(d2); 
  mySerial.print('.');
  mySerial.print(d3); 
  mySerial.print('.');
  mySerial.print(d4); 
  mySerial.print('.');
  mySerial.print('\r');
}



/*
 * Function that receives data from the ESP.
 * This is a possible breakpoint as errors can arise when using serial
 * communication at higher speeds.. We have to run a test which will ensure that
 * the values received are valid integers, and we can also check for the length of the received
 * data since we are expecting a specific number of characters from the ESP.
 */
void Receive_Data()
{
  if(mySerial.available())
  {
    String gh = ""; 
    gh = mySerial.readStringUntil('\r');

    Serial.print("Received Data from ESP: ");
    
    ltsensor = getValue(gh, '.', 1).toInt();
    rtsensor = getValue(gh, '.', 2).toInt();
    ldsensor = getValue(gh, '.', 3).toInt();
    rdsensor=  getValue(gh, '.', 4).toInt();
    
    Serial.print(gh); Serial.println();
    return ;
  }
}


/*
 * Function that splits a string into oomponent strings
 */
String getValue(String data, char separator, int index)
{
    int found = 0;
    int strIndex[] = { 0, -1 };
    int maxIndex = data.length() - 1;

    for (int i = 0; i <= maxIndex && found <= index; i++) {
        if (data.charAt(i) == separator || i == maxIndex) {
            found++;
            strIndex[0] = strIndex[1] + 1;
            strIndex[1] = (i == maxIndex) ? i+1 : i;
        }
    }
    return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}
